# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/variants.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nucleus.protos import reference_pb2 as nucleus_dot_protos_dot_reference__pb2
from nucleus.protos import struct_pb2 as nucleus_dot_protos_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/variants.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1dnucleus/protos/variants.proto\x12\x13nucleus.genomics.v1\x1a\x1enucleus/protos/reference.proto\x1a\x1bnucleus/protos/struct.proto\"\x9a\x03\n\x07Variant\x12\x16\n\x0ereference_name\x18\x0e \x01(\t\x12\r\n\x05start\x18\x10 \x01(\x03\x12\x0b\n\x03\x65nd\x18\r \x01(\x03\x12\r\n\x05names\x18\x03 \x03(\t\x12\x17\n\x0freference_bases\x18\x06 \x01(\t\x12\x17\n\x0f\x61lternate_bases\x18\x07 \x03(\t\x12\x0f\n\x07quality\x18\x08 \x01(\x01\x12\x0e\n\x06\x66ilter\x18\t \x03(\t\x12\x34\n\x04info\x18\n \x03(\x0b\x32&.nucleus.genomics.v1.Variant.InfoEntry\x12/\n\x05\x63\x61lls\x18\x0b \x03(\x0b\x32 .nucleus.genomics.v1.VariantCall\x12\x16\n\x0evariant_set_id\x18\x0f \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0f\n\x07\x63reated\x18\x0c \x01(\x03\x1aK\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.nucleus.genomics.v1.ListValue:\x02\x38\x01J\x04\x08\x01\x10\x02J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\"\xa6\x02\n\x0bVariantCall\x12\x15\n\rcall_set_name\x18\t \x01(\t\x12\x10\n\x08genotype\x18\x07 \x03(\x05\x12\x11\n\tis_phased\x18\n \x01(\x08\x12\x10\n\x08phaseset\x18\x05 \x01(\t\x12\x1b\n\x13genotype_likelihood\x18\x06 \x03(\x01\x12\x38\n\x04info\x18\x02 \x03(\x0b\x32*.nucleus.genomics.v1.VariantCall.InfoEntry\x12\x13\n\x0b\x63\x61ll_set_id\x18\x08 \x01(\t\x1aK\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.nucleus.genomics.v1.ListValue:\x02\x38\x01J\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\"\xf1\x02\n\tVcfHeader\x12\x12\n\nfileformat\x18\x01 \x01(\t\x12\x30\n\x07\x63ontigs\x18\x02 \x03(\x0b\x32\x1f.nucleus.genomics.v1.ContigInfo\x12\x33\n\x07\x66ilters\x18\x03 \x03(\x0b\x32\".nucleus.genomics.v1.VcfFilterInfo\x12+\n\x05infos\x18\x04 \x03(\x0b\x32\x1c.nucleus.genomics.v1.VcfInfo\x12\x33\n\x07\x66ormats\x18\x05 \x03(\x0b\x32\".nucleus.genomics.v1.VcfFormatInfo\x12\x14\n\x0csample_names\x18\x06 \x03(\t\x12\x42\n\x11structured_extras\x18\x08 \x03(\x0b\x32\'.nucleus.genomics.v1.VcfStructuredExtra\x12-\n\x06\x65xtras\x18\x07 \x03(\x0b\x32\x1d.nucleus.genomics.v1.VcfExtra\"0\n\rVcfFilterInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"i\n\x07VcfInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06number\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0e\n\x06source\x18\x05 \x01(\t\x12\x0f\n\x07version\x18\x06 \x01(\t\"N\n\rVcfFormatInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06number\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\"P\n\x12VcfStructuredExtra\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x06\x66ields\x18\x02 \x03(\x0b\x32\x1d.nucleus.genomics.v1.VcfExtra\"&\n\x08VcfExtra\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x81\x01\n\x10VcfReaderOptions\x12\x1c\n\x14\x65xcluded_info_fields\x18\x03 \x03(\t\x12\x1e\n\x16\x65xcluded_format_fields\x18\x04 \x03(\t\x12#\n\x1bstore_gl_and_pl_in_info_map\x18\x05 \x01(\x08J\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03\"\xcb\x01\n\x10VcfWriterOptions\x12\x1c\n\x14\x65xcluded_info_fields\x18\x07 \x03(\t\x12\x1e\n\x16\x65xcluded_format_fields\x18\x08 \x03(\t\x12\x19\n\x11round_qual_values\x18\x06 \x01(\x08\x12(\n retrieve_gl_and_pl_from_info_map\x18\t \x01(\x08\x12\x16\n\x0e\x65xclude_header\x18\n \x01(\x08J\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\x62\x06proto3')
  ,
  dependencies=[nucleus_dot_protos_dot_reference__pb2.DESCRIPTOR,nucleus_dot_protos_dot_struct__pb2.DESCRIPTOR,])




_VARIANT_INFOENTRY = _descriptor.Descriptor(
  name='InfoEntry',
  full_name='nucleus.genomics.v1.Variant.InfoEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.Variant.InfoEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.Variant.InfoEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=433,
  serialized_end=508,
)

_VARIANT = _descriptor.Descriptor(
  name='Variant',
  full_name='nucleus.genomics.v1.Variant',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reference_name', full_name='nucleus.genomics.v1.Variant.reference_name', index=0,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start', full_name='nucleus.genomics.v1.Variant.start', index=1,
      number=16, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end', full_name='nucleus.genomics.v1.Variant.end', index=2,
      number=13, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='names', full_name='nucleus.genomics.v1.Variant.names', index=3,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reference_bases', full_name='nucleus.genomics.v1.Variant.reference_bases', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alternate_bases', full_name='nucleus.genomics.v1.Variant.alternate_bases', index=5,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quality', full_name='nucleus.genomics.v1.Variant.quality', index=6,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filter', full_name='nucleus.genomics.v1.Variant.filter', index=7,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='info', full_name='nucleus.genomics.v1.Variant.info', index=8,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='calls', full_name='nucleus.genomics.v1.Variant.calls', index=9,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variant_set_id', full_name='nucleus.genomics.v1.Variant.variant_set_id', index=10,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.Variant.id', index=11,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='created', full_name='nucleus.genomics.v1.Variant.created', index=12,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_VARIANT_INFOENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=116,
  serialized_end=526,
)


_VARIANTCALL_INFOENTRY = _descriptor.Descriptor(
  name='InfoEntry',
  full_name='nucleus.genomics.v1.VariantCall.InfoEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.VariantCall.InfoEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.VariantCall.InfoEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=433,
  serialized_end=508,
)

_VARIANTCALL = _descriptor.Descriptor(
  name='VariantCall',
  full_name='nucleus.genomics.v1.VariantCall',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='call_set_name', full_name='nucleus.genomics.v1.VariantCall.call_set_name', index=0,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='genotype', full_name='nucleus.genomics.v1.VariantCall.genotype', index=1,
      number=7, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_phased', full_name='nucleus.genomics.v1.VariantCall.is_phased', index=2,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='phaseset', full_name='nucleus.genomics.v1.VariantCall.phaseset', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='genotype_likelihood', full_name='nucleus.genomics.v1.VariantCall.genotype_likelihood', index=4,
      number=6, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='info', full_name='nucleus.genomics.v1.VariantCall.info', index=5,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='call_set_id', full_name='nucleus.genomics.v1.VariantCall.call_set_id', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_VARIANTCALL_INFOENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=529,
  serialized_end=823,
)


_VCFHEADER = _descriptor.Descriptor(
  name='VcfHeader',
  full_name='nucleus.genomics.v1.VcfHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='fileformat', full_name='nucleus.genomics.v1.VcfHeader.fileformat', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contigs', full_name='nucleus.genomics.v1.VcfHeader.contigs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filters', full_name='nucleus.genomics.v1.VcfHeader.filters', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='infos', full_name='nucleus.genomics.v1.VcfHeader.infos', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='formats', full_name='nucleus.genomics.v1.VcfHeader.formats', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_names', full_name='nucleus.genomics.v1.VcfHeader.sample_names', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='structured_extras', full_name='nucleus.genomics.v1.VcfHeader.structured_extras', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extras', full_name='nucleus.genomics.v1.VcfHeader.extras', index=7,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=1195,
)


_VCFFILTERINFO = _descriptor.Descriptor(
  name='VcfFilterInfo',
  full_name='nucleus.genomics.v1.VcfFilterInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.VcfFilterInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.VcfFilterInfo.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1197,
  serialized_end=1245,
)


_VCFINFO = _descriptor.Descriptor(
  name='VcfInfo',
  full_name='nucleus.genomics.v1.VcfInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.VcfInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number', full_name='nucleus.genomics.v1.VcfInfo.number', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='nucleus.genomics.v1.VcfInfo.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.VcfInfo.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source', full_name='nucleus.genomics.v1.VcfInfo.source', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='nucleus.genomics.v1.VcfInfo.version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1247,
  serialized_end=1352,
)


_VCFFORMATINFO = _descriptor.Descriptor(
  name='VcfFormatInfo',
  full_name='nucleus.genomics.v1.VcfFormatInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.VcfFormatInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number', full_name='nucleus.genomics.v1.VcfFormatInfo.number', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='nucleus.genomics.v1.VcfFormatInfo.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.VcfFormatInfo.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1354,
  serialized_end=1432,
)


_VCFSTRUCTUREDEXTRA = _descriptor.Descriptor(
  name='VcfStructuredExtra',
  full_name='nucleus.genomics.v1.VcfStructuredExtra',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.VcfStructuredExtra.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fields', full_name='nucleus.genomics.v1.VcfStructuredExtra.fields', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1434,
  serialized_end=1514,
)


_VCFEXTRA = _descriptor.Descriptor(
  name='VcfExtra',
  full_name='nucleus.genomics.v1.VcfExtra',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.VcfExtra.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.VcfExtra.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1516,
  serialized_end=1554,
)


_VCFREADEROPTIONS = _descriptor.Descriptor(
  name='VcfReaderOptions',
  full_name='nucleus.genomics.v1.VcfReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='excluded_info_fields', full_name='nucleus.genomics.v1.VcfReaderOptions.excluded_info_fields', index=0,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='excluded_format_fields', full_name='nucleus.genomics.v1.VcfReaderOptions.excluded_format_fields', index=1,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='store_gl_and_pl_in_info_map', full_name='nucleus.genomics.v1.VcfReaderOptions.store_gl_and_pl_in_info_map', index=2,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1557,
  serialized_end=1686,
)


_VCFWRITEROPTIONS = _descriptor.Descriptor(
  name='VcfWriterOptions',
  full_name='nucleus.genomics.v1.VcfWriterOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='excluded_info_fields', full_name='nucleus.genomics.v1.VcfWriterOptions.excluded_info_fields', index=0,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='excluded_format_fields', full_name='nucleus.genomics.v1.VcfWriterOptions.excluded_format_fields', index=1,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='round_qual_values', full_name='nucleus.genomics.v1.VcfWriterOptions.round_qual_values', index=2,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='retrieve_gl_and_pl_from_info_map', full_name='nucleus.genomics.v1.VcfWriterOptions.retrieve_gl_and_pl_from_info_map', index=3,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exclude_header', full_name='nucleus.genomics.v1.VcfWriterOptions.exclude_header', index=4,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1689,
  serialized_end=1892,
)

_VARIANT_INFOENTRY.fields_by_name['value'].message_type = nucleus_dot_protos_dot_struct__pb2._LISTVALUE
_VARIANT_INFOENTRY.containing_type = _VARIANT
_VARIANT.fields_by_name['info'].message_type = _VARIANT_INFOENTRY
_VARIANT.fields_by_name['calls'].message_type = _VARIANTCALL
_VARIANTCALL_INFOENTRY.fields_by_name['value'].message_type = nucleus_dot_protos_dot_struct__pb2._LISTVALUE
_VARIANTCALL_INFOENTRY.containing_type = _VARIANTCALL
_VARIANTCALL.fields_by_name['info'].message_type = _VARIANTCALL_INFOENTRY
_VCFHEADER.fields_by_name['contigs'].message_type = nucleus_dot_protos_dot_reference__pb2._CONTIGINFO
_VCFHEADER.fields_by_name['filters'].message_type = _VCFFILTERINFO
_VCFHEADER.fields_by_name['infos'].message_type = _VCFINFO
_VCFHEADER.fields_by_name['formats'].message_type = _VCFFORMATINFO
_VCFHEADER.fields_by_name['structured_extras'].message_type = _VCFSTRUCTUREDEXTRA
_VCFHEADER.fields_by_name['extras'].message_type = _VCFEXTRA
_VCFSTRUCTUREDEXTRA.fields_by_name['fields'].message_type = _VCFEXTRA
DESCRIPTOR.message_types_by_name['Variant'] = _VARIANT
DESCRIPTOR.message_types_by_name['VariantCall'] = _VARIANTCALL
DESCRIPTOR.message_types_by_name['VcfHeader'] = _VCFHEADER
DESCRIPTOR.message_types_by_name['VcfFilterInfo'] = _VCFFILTERINFO
DESCRIPTOR.message_types_by_name['VcfInfo'] = _VCFINFO
DESCRIPTOR.message_types_by_name['VcfFormatInfo'] = _VCFFORMATINFO
DESCRIPTOR.message_types_by_name['VcfStructuredExtra'] = _VCFSTRUCTUREDEXTRA
DESCRIPTOR.message_types_by_name['VcfExtra'] = _VCFEXTRA
DESCRIPTOR.message_types_by_name['VcfReaderOptions'] = _VCFREADEROPTIONS
DESCRIPTOR.message_types_by_name['VcfWriterOptions'] = _VCFWRITEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Variant = _reflection.GeneratedProtocolMessageType('Variant', (_message.Message,), {

  'InfoEntry' : _reflection.GeneratedProtocolMessageType('InfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_INFOENTRY,
    '__module__' : 'nucleus.protos.variants_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Variant.InfoEntry)
    })
  ,
  'DESCRIPTOR' : _VARIANT,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.Variant)
  })
_sym_db.RegisterMessage(Variant)
_sym_db.RegisterMessage(Variant.InfoEntry)

VariantCall = _reflection.GeneratedProtocolMessageType('VariantCall', (_message.Message,), {

  'InfoEntry' : _reflection.GeneratedProtocolMessageType('InfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _VARIANTCALL_INFOENTRY,
    '__module__' : 'nucleus.protos.variants_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VariantCall.InfoEntry)
    })
  ,
  'DESCRIPTOR' : _VARIANTCALL,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VariantCall)
  })
_sym_db.RegisterMessage(VariantCall)
_sym_db.RegisterMessage(VariantCall.InfoEntry)

VcfHeader = _reflection.GeneratedProtocolMessageType('VcfHeader', (_message.Message,), {
  'DESCRIPTOR' : _VCFHEADER,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfHeader)
  })
_sym_db.RegisterMessage(VcfHeader)

VcfFilterInfo = _reflection.GeneratedProtocolMessageType('VcfFilterInfo', (_message.Message,), {
  'DESCRIPTOR' : _VCFFILTERINFO,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfFilterInfo)
  })
_sym_db.RegisterMessage(VcfFilterInfo)

VcfInfo = _reflection.GeneratedProtocolMessageType('VcfInfo', (_message.Message,), {
  'DESCRIPTOR' : _VCFINFO,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfInfo)
  })
_sym_db.RegisterMessage(VcfInfo)

VcfFormatInfo = _reflection.GeneratedProtocolMessageType('VcfFormatInfo', (_message.Message,), {
  'DESCRIPTOR' : _VCFFORMATINFO,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfFormatInfo)
  })
_sym_db.RegisterMessage(VcfFormatInfo)

VcfStructuredExtra = _reflection.GeneratedProtocolMessageType('VcfStructuredExtra', (_message.Message,), {
  'DESCRIPTOR' : _VCFSTRUCTUREDEXTRA,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfStructuredExtra)
  })
_sym_db.RegisterMessage(VcfStructuredExtra)

VcfExtra = _reflection.GeneratedProtocolMessageType('VcfExtra', (_message.Message,), {
  'DESCRIPTOR' : _VCFEXTRA,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfExtra)
  })
_sym_db.RegisterMessage(VcfExtra)

VcfReaderOptions = _reflection.GeneratedProtocolMessageType('VcfReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _VCFREADEROPTIONS,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfReaderOptions)
  })
_sym_db.RegisterMessage(VcfReaderOptions)

VcfWriterOptions = _reflection.GeneratedProtocolMessageType('VcfWriterOptions', (_message.Message,), {
  'DESCRIPTOR' : _VCFWRITEROPTIONS,
  '__module__' : 'nucleus.protos.variants_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.VcfWriterOptions)
  })
_sym_db.RegisterMessage(VcfWriterOptions)


_VARIANT_INFOENTRY._options = None
_VARIANTCALL_INFOENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Struct proto utilities.

This class provides wrappers for conveniently interacting with protos defined
in struct.proto, mostly ListValue and Value objects. It should primarily be used
by variant_utils and variantcallutils rather than being used directly.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import types

from nucleus.protos import struct_pb2

# Field names of values defined in struct_pb2.Value.
_BOOL_TYPE = 'bool_value'
_INT_TYPE = 'int_value'
_NUMBER_TYPE = 'number_value'
_STRING_TYPE = 'string_value'


def _add_field_with_type(field_map, field_name, value, value_type):
  """Adds values to a particular map field containing a ListValue."""
  if not isinstance(value, (list, types.GeneratorType, tuple)):
    value = [value]
  struct_values = [struct_pb2.Value(**{value_type: v}) for v in value]
  field_map[field_name].values.extend(struct_values)


def _set_field_with_type(field_map, field_name, value, value_type):
  """Sets values to a particular map field containing a ListValue."""
  if field_name in field_map:
    del field_map[field_name]
  _add_field_with_type(field_map, field_name, value, value_type)


def _get_field_with_type(field_map, field_name, is_single_field, value_type):
  fields = [getattr(v, value_type) for v in field_map[field_name].values]
  return fields[0] if is_single_field and fields else fields


def add_number_field(field_map, field_name, value):
  """Appends the given number value(s) to field_map[field_name].

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to append value to.
    value: The number value(s) to append to the field. This can be a single
      number or a list of numbers.
  """
  _add_field_with_type(field_map, field_name, value, _NUMBER_TYPE)


def set_number_field(field_map, field_name, value):
  """Sets field_map[field_name] with the given number value(s).

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to set.
    value: The number value(s) to set the field to. This can be a single number
      or a list of numbers.
  """
  _set_field_with_type(field_map, field_name, value, _NUMBER_TYPE)


def get_number_field(field_map, field_name, is_single_field=False):
  """Returns the number value(s) stored in `field_map[field_name]`.

  If the field_name is not present in field_map, the empty list is returned.

  Args:
    field_map: Map(str --> ListValue) of interest.
    field_name: str. The name of the field to extract number values from.
    is_single_field: bool. If True, return the first number value stored (it
      should be the only one in the field). Otherwise, return the list of
      numbers.

  Returns:
    The number value(s) stored in the field_map under this field_name.
  """
  return _get_field_with_type(field_map, field_name, is_single_field,
                              _NUMBER_TYPE)


def add_int_field(field_map, field_name, value):
  """Appends the given int value(s) to field_map[field_name].

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to append value to.
    value: The int value(s) to append to the field. This can be a single
      int or a list of ints.
  """
  _add_field_with_type(field_map, field_name, value, _INT_TYPE)


def set_int_field(field_map, field_name, value):
  """Sets field_map[field_name] with the given int value(s).

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to set.
    value: The int value(s) to set the field to. This can be a single int
      or a list of ints.
  """
  _set_field_with_type(field_map, field_name, value, _INT_TYPE)


def get_int_field(field_map, field_name, is_single_field=False):
  """Returns the int value(s) stored in `field_map[field_name]`.

  If the field_name is not present in field_map, the empty list is returned.

  Args:
    field_map: Map(str --> ListValue) of interest.
    field_name: str. The name of the field to extract int values from.
    is_single_field: bool. If True, return the first int value stored (it
      should be the only one in the field). Otherwise, return the list of
      ints.

  Returns:
    The int value(s) stored in the field_map under this field_name.
  """
  return _get_field_with_type(field_map, field_name, is_single_field, _INT_TYPE)


def add_string_field(field_map, field_name, value):
  """Appends the given string value(s) to field_map[field_name].

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to append value to.
    value: The string value(s) to append to the field. This can be a single
      string or a list of strings.
  """
  _add_field_with_type(field_map, field_name, value, _STRING_TYPE)


def set_string_field(field_map, field_name, value):
  """Sets field_map[field_name] with the given string value(s).

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to set.
    value: The int value(s) to set the field to. This can be a single string or
      a list of strings.
  """
  _set_field_with_type(field_map, field_name, value, _STRING_TYPE)


def get_string_field(field_map, field_name, is_single_field=False):
  """Returns the string value(s) stored in `field_map[field_name]`.

  If the field_name is not present in field_map, the empty list is returned.

  Args:
    field_map: Map(str --> ListValue) of interest.
    field_name: str. The name of the field to extract string values from.
    is_single_field: bool. If True, return the first string value stored (it
      should be the only one in the field). Otherwise, return the list of
      strings.

  Returns:
    The string value(s) stored in the field_map under this field_name.
  """
  return _get_field_with_type(field_map, field_name, is_single_field,
                              _STRING_TYPE)


def add_bool_field(field_map, field_name, value):
  """Appends the given boolean value(s) to field_map[field_name].

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to append value to.
    value: The boolean value(s) to append to the field. This can be a single
      boolean or a list of booleans.
  """
  _add_field_with_type(field_map, field_name, value, _BOOL_TYPE)


def set_bool_field(field_map, field_name, value):
  """Sets field_map[field_name] with the given boolean value(s).

  Args:
    field_map: Map(str --> ListValue) to modify.
    field_name: str. The name of the field to set.
    value: The boolean value(s) to set the field to. This can be a single
      boolean or a list of booleans.
  """
  _set_field_with_type(field_map, field_name, value, _BOOL_TYPE)


def get_bool_field(field_map, field_name, is_single_field=False):
  """Returns the bool value(s) stored in `field_map[field_name]`.

  If the field_name is not present in field_map, the empty list is returned.

  Args:
    field_map: Map(str --> ListValue) of interest.
    field_name: str. The name of the field to extract bool values from.
    is_single_field: bool. If True, return the first bool value stored (it
      should be the only one in the field). Otherwise, return the list of
      bools.

  Returns:
    The bool value(s) stored in the field_map under this field_name.
  """
  return _get_field_with_type(field_map, field_name, is_single_field,
                              _BOOL_TYPE)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants related to the VCF variant specification.

See the full specification at https://samtools.github.io/hts-specs/VCFv4.3.pdf
for details.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

from nucleus.protos import variants_pb2
from nucleus.util import struct_utils

# The alternate allele string for reference (no alt).
NO_ALT_ALLELE = '.'

# The alternate allele string for the gVCF "any" alternate allele.
GVCF_ALT_ALLELE = '<*>'

# Older symbolic alt allele, similar in meaning to gVCF alt allele
SYMBOLIC_ALT_ALLELE = '<NON_REF>'

# The replacement field used for missing data.
MISSING_FIELD = '.'

# Valid types for INFO and FORMAT fields, as per the VCF 4.3 spec.
CHARACTER_TYPE = 'Character'
FLAG_TYPE = 'Flag'
FLOAT_TYPE = 'Float'
INTEGER_TYPE = 'Integer'
STRING_TYPE = 'String'

# Reserved FILTER field definitions.
RESERVED_FILTER_FIELDS = [
    variants_pb2.VcfFilterInfo(id='PASS', description='All filters passed'),
]

# Reserved INFO field definitions, as per the VCF 4.3 spec.
RESERVED_INFO_FIELDS = [
    variants_pb2.VcfInfo(
        id='AA', number='1', type=STRING_TYPE, description='Ancestral allele'),
    variants_pb2.VcfInfo(
        id='AC',
        number='A',
        type=INTEGER_TYPE,
        description='Allele count in genotypes, for each ALT '
        'allele, in the same order as listed'),
    variants_pb2.VcfInfo(
        id='AD',
        number='R',
        type=INTEGER_TYPE,
        description='Total read depth for each allele'),
    variants_pb2.VcfInfo(
        id='ADF',
        number='R',
        type=INTEGER_TYPE,
        description='Read depth for each allele on the forward '
        'strand'),
    variants_pb2.VcfInfo(
        id='ADR',
        number='R',
        type=INTEGER_TYPE,
        description='Read depth for each allele on the reverse strand'),
    variants_pb2.VcfInfo(
        id='AF',
        number='A',
        type=FLOAT_TYPE,
        description='Allele frequency for each ALT allele in '
        'the same order as listed (estimated from '
        'primary data, not called genotypes)'),
    variants_pb2.VcfInfo(
        id='AN',
        number='1',
        type=INTEGER_TYPE,
        description='Total number of alleles in called genotypes'),
    variants_pb2.VcfInfo(
        id='BQ', number='1', type=FLOAT_TYPE, description='RMS base quality'),
    variants_pb2.VcfInfo(
        id='CIGAR',
        number='A',
        type=STRING_TYPE,
        description='Cigar string describing how to align an '
        'alternate allele to the reference allele'),
    variants_pb2.VcfInfo(
        id='DB', number='0', type=FLAG_TYPE, description='dbSNP membership'),
    variants_pb2.VcfInfo(
        id='DP',
        number='1',
        type=INTEGER_TYPE,
        description='Combined depth across samples'),
    variants_pb2.VcfInfo(
        id='END',
        number='1',
        type=INTEGER_TYPE,
        description='End position (for use with symbolic alleles)'),
    variants_pb2.VcfInfo(
        id='H2', number='0', type=FLAG_TYPE, description='HapMap2 membership'),
    variants_pb2.VcfInfo(
        id='H3', number='0', type=FLAG_TYPE, description='HapMap3 membership'),
    # NOTE: In the VCF 4.3 spec, the type of 'MQ' is listed as '.', even though
    # that is not specified as a valid type. Because root mean square is
    # typically a float value, we specify its type as FLOAT_TYPE.
    variants_pb2.VcfInfo(
        id='MQ', number='1', type=FLOAT_TYPE,
        description='RMS mapping quality'),
    variants_pb2.VcfInfo(
        id='MQ0',
        number='1',
        type=INTEGER_TYPE,
        description='Number of MAPQ == 0 reads'),
    variants_pb2.VcfInfo(
        id='NS',
        number='1',
        type=INTEGER_TYPE,
        description='Number of samples with data'),
    # NOTE: In the VCF 4.3 spec, the type of 'SB' is listed as '.', even though
    # that is not specified as a valid type. Because strand bias is usually a
    # numerical measurement (e.g. p-value of contingency table), we specify its
    # type as FLOAT_TYPE.
    variants_pb2.VcfInfo(
        id='SB', number='.', type=FLOAT_TYPE, description='Strand bias'),
    variants_pb2.VcfInfo(
        id='SOMATIC',
        number='0',
        type=FLAG_TYPE,
        description='Somatic mutation (for cancer genomics)'),
    variants_pb2.VcfInfo(
        id='VALIDATED',
        number='0',
        type=FLAG_TYPE,
        description='Validated by follow-up experiment'),
    variants_pb2.VcfInfo(
        id='1000G',
        number='0',
        type=FLAG_TYPE,
        description='1000 Genomes membership'),
]

# Reserved FORMAT field definitions, as per the VCF 4.3 spec.
RESERVED_FORMAT_FIELDS = [
    variants_pb2.VcfFormatInfo(
        id='AD',
        number='R',
        type=INTEGER_TYPE,
        description='Read depth for each allele'),
    variants_pb2.VcfFormatInfo(
        id='ADF',
        number='R',
        type=INTEGER_TYPE,
        description='Read depth for each allele on the '
        'forward strand'),
    variants_pb2.VcfFormatInfo(
        id='ADR',
        number='R',
        type=INTEGER_TYPE,
        description='Read depth for each allele on the '
        'reverse strand'),
    variants_pb2.VcfFormatInfo(
        id='DP', number='1', type=INTEGER_TYPE, description='Read depth'),
    variants_pb2.VcfFormatInfo(
        id='EC',
        number='A',
        type=INTEGER_TYPE,
        description='Expected alternate allele counts'),
    variants_pb2.VcfFormatInfo(
        id='FT',
        number='1',
        type=STRING_TYPE,
        description='Filter indicating if this genotype '
        'was "called"'),
    variants_pb2.VcfFormatInfo(
        id='GL',
        number='G',
        type=FLOAT_TYPE,
        description='Genotype likelihoods'),
    variants_pb2.VcfFormatInfo(
        id='GP',
        number='G',
        type=FLOAT_TYPE,
        description='Genotype posterior probabilities'),
    variants_pb2.VcfFormatInfo(
        id='GQ',
        number='1',
        type=INTEGER_TYPE,
        description='Conditional genotype quality'),
    variants_pb2.VcfFormatInfo(
        id='GT', number='1', type=STRING_TYPE, description='Genotype'),
    variants_pb2.VcfFormatInfo(
        id='HQ', number='2', type=INTEGER_TYPE,
        description='Haplotype quality'),
    variants_pb2.VcfFormatInfo(
        id='MQ',
        number='1',
        type=INTEGER_TYPE,
        description='RMS mapping quality'),
    variants_pb2.VcfFormatInfo(
        id='PL',
        number='G',
        type=INTEGER_TYPE,
        description='Phred-scaled genotype likelihoods '
        'rounded to the closest integer'),
    variants_pb2.VcfFormatInfo(
        id='PQ', number='1', type=INTEGER_TYPE, description='Phasing quality'),
    variants_pb2.VcfFormatInfo(
        id='PS', number='1', type=INTEGER_TYPE, description='Phase set'),
]

# Map from field type to the function used to set struct_pb2.Value elements
# of that type.
SET_FN_LOOKUP = {
    INTEGER_TYPE: struct_utils.set_int_field,
    FLOAT_TYPE: struct_utils.set_number_field,
    STRING_TYPE: struct_utils.set_string_field,
    CHARACTER_TYPE: struct_utils.set_string_field,
    FLAG_TYPE: struct_utils.set_bool_field,
}


def _get_reserved_field(field_id, reserved_fields):
  """Returns the desired reserved field.

  Args:
    field_id: str. The id of the field to retrieve.
    reserved_fields: list(fields). The reserved fields to search.

  Returns:
    The reserved field with the given `field_id`.

  Raises:
    ValueError: `field_id` is not a known reserved field.
  """
  matching_fields = [field for field in reserved_fields if field.id == field_id]
  if not matching_fields:
    raise ValueError('No reserved field with id `{}`'.format(field_id))
  return matching_fields[0]


def reserved_filter_field(field_id):
  """Returns the reserved FILTER field with the given ID."""
  return _get_reserved_field(field_id, RESERVED_FILTER_FIELDS)


def reserved_info_field(field_id):
  """Returns the reserved INFO field with the given ID."""
  return _get_reserved_field(field_id, RESERVED_INFO_FIELDS)


def reserved_format_field(field_id):
  """Returns the reserved FORMAT field with the given ID."""
  return _get_reserved_field(field_id, RESERVED_FORMAT_FIELDS)


def create_get_fn(value_type, number):
  """Returns a callable that extracts the typed information from a ListValue.

  Args:
    value_type: str. The value type stored as defined in the VCF 4.3 spec.
    number: str. The number of entries of this value as defined in the VCF spec.

  Returns:
    A callable that takes two inputs: A Map(str --> ListValue) and a string
    field name and returns the associated typed value(s). The return value is
    a list of typed values or a single typed value, depending on the expected
    number of values returned.
  """
  is_single_field = (number == '0' or number == '1')
  if value_type == CHARACTER_TYPE or value_type == STRING_TYPE:
    return functools.partial(
        struct_utils.get_string_field, is_single_field=is_single_field)
  elif value_type == INTEGER_TYPE:
    return functools.partial(
        struct_utils.get_int_field, is_single_field=is_single_field)
  elif value_type == FLOAT_TYPE:
    return functools.partial(
        struct_utils.get_number_field, is_single_field=is_single_field)
  elif value_type == FLAG_TYPE:
    return functools.partial(
        struct_utils.get_bool_field, is_single_field=is_single_field)
  else:
    raise ValueError('Invalid value_type: {}'.format(value_type))


# Map from INFO field name to the function used to set struct_pb2.Value elements
# of that field.
RESERVED_INFO_FIELD_SET_FNS = {
    info.id: SET_FN_LOOKUP[info.type]
    for info in RESERVED_INFO_FIELDS
}

# Map from INFO field name to the function used to get struct_pb2.Value elements
# of that field.
RESERVED_INFO_FIELD_GET_FNS = {
    info.id: create_get_fn(info.type, info.number)
    for info in RESERVED_INFO_FIELDS
}

# Map from FORMAT field name to the function used to set struct_pb2.Value
# elements of that field.
RESERVED_FORMAT_FIELD_SET_FNS = {
    fmt.id: SET_FN_LOOKUP[fmt.type]
    for fmt in RESERVED_FORMAT_FIELDS
}

# Map from FORMAT field name to the function used to get struct_pb2.Value
# elements of that field.
RESERVED_FORMAT_FIELD_GET_FNS = {
    fmt.id: create_get_fn(fmt.type, fmt.number)
    for fmt in RESERVED_FORMAT_FIELDS
}


def reserved_info_field_set_fn(field_name):
  """Returns the callable that sets the proper field for the given field_name.

  Args:
    field_name: str. The field name of the reserved INFO field (e.g. 'MQ').

  Returns:
    The callable that takes in a Map(str --> ListValue), field name, and value
    and modifies the map to populate the field_name entry with the given value.

  Raises:
    ValueError: The field_name is not a known reserved INFO field.
  """
  try:
    return RESERVED_INFO_FIELD_SET_FNS[field_name]
  except KeyError:
    raise ValueError('Unknown reserved INFO field: {}'.format(field_name))


def reserved_info_field_get_fn(field_name):
  """Returns the callable that gets the proper field for the given field_name.

  Args:
    field_name: str. The field name of the reserved INFO field (e.g. 'MQ').

  Returns:
    The callable that takes in a Map(str --> ListValue), and field name and
    returns the associated typed value(s).

  Raises:
    ValueError: The field_name is not a known reserved INFO field.
  """
  try:
    return RESERVED_INFO_FIELD_GET_FNS[field_name]
  except KeyError:
    raise ValueError(
        'Unknown reserved INFO field to get: {}'.format(field_name))


def reserved_format_field_set_fn(field_name):
  """Returns the callable that sets the proper field for the given field_name.

  Args:
    field_name: str. The field name of the reserved FORMAT field (e.g. 'AD').

  Returns:
    The callable that takes in a Map(str --> ListValue), field name, and value
    and modifies the map to populate the field_name entry with the given value.

  Raises:
    ValueError: The field_name is not a known reserved FORMAT field.
  """
  try:
    return RESERVED_FORMAT_FIELD_SET_FNS[field_name]
  except KeyError:
    raise ValueError('Unknown reserved FORMAT field: {}'.format(field_name))


def reserved_format_field_get_fn(field_name):
  """Returns the callable that gets the proper field for the given field_name.

  Args:
    field_name: str. The field name of the reserved FORMAT field (e.g. 'AD').

  Returns:
    The callable that takes in a Map(str --> ListValue), and field name and
    returns the associated typed value(s).

  Raises:
    ValueError: The field_name is not a known reserved FORMAT field.
  """
  try:
    return RESERVED_FORMAT_FIELD_GET_FNS[field_name]
  except KeyError:
    raise ValueError(
        'Unknown reserved FORMAT field to get: {}'.format(field_name))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/bed.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/bed.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x18nucleus/protos/bed.proto\x12\x13nucleus.genomics.v1\"\xce\x02\n\tBedRecord\x12\x16\n\x0ereference_name\x18\x01 \x01(\t\x12\r\n\x05start\x18\x02 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x03\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\r\n\x05score\x18\x05 \x01(\x01\x12\x35\n\x06strand\x18\x06 \x01(\x0e\x32%.nucleus.genomics.v1.BedRecord.Strand\x12\x13\n\x0bthick_start\x18\x07 \x01(\x03\x12\x11\n\tthick_end\x18\x08 \x01(\x03\x12\x10\n\x08item_rgb\x18\t \x01(\t\x12\x13\n\x0b\x62lock_count\x18\n \x01(\x05\x12\x13\n\x0b\x62lock_sizes\x18\x0b \x01(\t\x12\x14\n\x0c\x62lock_starts\x18\x0c \x01(\t\"?\n\x06Strand\x12\r\n\tNO_STRAND\x10\x00\x12\x12\n\x0e\x46ORWARD_STRAND\x10\x01\x12\x12\n\x0eREVERSE_STRAND\x10\x02\"\x1f\n\tBedHeader\x12\x12\n\nnum_fields\x18\x01 \x01(\x05\",\n\x10\x42\x65\x64ReaderOptions\x12\x12\n\nnum_fields\x18\x02 \x01(\x05J\x04\x08\x01\x10\x02\"\x12\n\x10\x42\x65\x64WriterOptionsb\x06proto3')
)



_BEDRECORD_STRAND = _descriptor.EnumDescriptor(
  name='Strand',
  full_name='nucleus.genomics.v1.BedRecord.Strand',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NO_STRAND', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORWARD_STRAND', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REVERSE_STRAND', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=321,
  serialized_end=384,
)
_sym_db.RegisterEnumDescriptor(_BEDRECORD_STRAND)


_BEDRECORD = _descriptor.Descriptor(
  name='BedRecord',
  full_name='nucleus.genomics.v1.BedRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reference_name', full_name='nucleus.genomics.v1.BedRecord.reference_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start', full_name='nucleus.genomics.v1.BedRecord.start', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end', full_name='nucleus.genomics.v1.BedRecord.end', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='nucleus.genomics.v1.BedRecord.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='score', full_name='nucleus.genomics.v1.BedRecord.score', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='strand', full_name='nucleus.genomics.v1.BedRecord.strand', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='thick_start', full_name='nucleus.genomics.v1.BedRecord.thick_start', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='thick_end', full_name='nucleus.genomics.v1.BedRecord.thick_end', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='item_rgb', full_name='nucleus.genomics.v1.BedRecord.item_rgb', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_count', full_name='nucleus.genomics.v1.BedRecord.block_count', index=9,
      number=10, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_sizes', full_name='nucleus.genomics.v1.BedRecord.block_sizes', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_starts', full_name='nucleus.genomics.v1.BedRecord.block_starts', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BEDRECORD_STRAND,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=384,
)


_BEDHEADER = _descriptor.Descriptor(
  name='BedHeader',
  full_name='nucleus.genomics.v1.BedHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_fields', full_name='nucleus.genomics.v1.BedHeader.num_fields', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=386,
  serialized_end=417,
)


_BEDREADEROPTIONS = _descriptor.Descriptor(
  name='BedReaderOptions',
  full_name='nucleus.genomics.v1.BedReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_fields', full_name='nucleus.genomics.v1.BedReaderOptions.num_fields', index=0,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=419,
  serialized_end=463,
)


_BEDWRITEROPTIONS = _descriptor.Descriptor(
  name='BedWriterOptions',
  full_name='nucleus.genomics.v1.BedWriterOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=465,
  serialized_end=483,
)

_BEDRECORD.fields_by_name['strand'].enum_type = _BEDRECORD_STRAND
_BEDRECORD_STRAND.containing_type = _BEDRECORD
DESCRIPTOR.message_types_by_name['BedRecord'] = _BEDRECORD
DESCRIPTOR.message_types_by_name['BedHeader'] = _BEDHEADER
DESCRIPTOR.message_types_by_name['BedReaderOptions'] = _BEDREADEROPTIONS
DESCRIPTOR.message_types_by_name['BedWriterOptions'] = _BEDWRITEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BedRecord = _reflection.GeneratedProtocolMessageType('BedRecord', (_message.Message,), {
  'DESCRIPTOR' : _BEDRECORD,
  '__module__' : 'nucleus.protos.bed_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.BedRecord)
  })
_sym_db.RegisterMessage(BedRecord)

BedHeader = _reflection.GeneratedProtocolMessageType('BedHeader', (_message.Message,), {
  'DESCRIPTOR' : _BEDHEADER,
  '__module__' : 'nucleus.protos.bed_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.BedHeader)
  })
_sym_db.RegisterMessage(BedHeader)

BedReaderOptions = _reflection.GeneratedProtocolMessageType('BedReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _BEDREADEROPTIONS,
  '__module__' : 'nucleus.protos.bed_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.BedReaderOptions)
  })
_sym_db.RegisterMessage(BedReaderOptions)

BedWriterOptions = _reflection.GeneratedProtocolMessageType('BedWriterOptions', (_message.Message,), {
  'DESCRIPTOR' : _BEDWRITEROPTIONS,
  '__module__' : 'nucleus.protos.bed_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.BedWriterOptions)
  })
_sym_db.RegisterMessage(BedWriterOptions)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/gff.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nucleus.protos import range_pb2 as nucleus_dot_protos_dot_range__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/gff.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x18nucleus/protos/gff.proto\x12\x13nucleus.genomics.v1\x1a\x1anucleus/protos/range.proto\"\xea\x02\n\tGffRecord\x12)\n\x05range\x18\x01 \x01(\x0b\x32\x1a.nucleus.genomics.v1.Range\x12\x0e\n\x06source\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\r\n\x05score\x18\x04 \x01(\x01\x12\x35\n\x06strand\x18\x05 \x01(\x0e\x32%.nucleus.genomics.v1.GffRecord.Strand\x12\r\n\x05phase\x18\x06 \x01(\x05\x12\x42\n\nattributes\x18\x07 \x03(\x0b\x32..nucleus.genomics.v1.GffRecord.AttributesEntry\x1a\x31\n\x0f\x41ttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"H\n\x06Strand\x12\x16\n\x12UNSPECIFIED_STRAND\x10\x00\x12\x12\n\x0e\x46ORWARD_STRAND\x10\x01\x12\x12\n\x0eREVERSE_STRAND\x10\x02\"\xf5\x03\n\tGffHeader\x12\x13\n\x0bgff_version\x18\x01 \x01(\t\x12\x34\n\x10sequence_regions\x18\x02 \x03(\x0b\x32\x1a.nucleus.genomics.v1.Range\x12L\n\x12\x66\x65\x61ture_ontologies\x18\x03 \x03(\x0b\x32\x30.nucleus.genomics.v1.GffHeader.OntologyDirective\x12N\n\x14\x61ttribute_ontologies\x18\x04 \x03(\x0b\x32\x30.nucleus.genomics.v1.GffHeader.OntologyDirective\x12K\n\x11source_ontologies\x18\x05 \x03(\x0b\x32\x30.nucleus.genomics.v1.GffHeader.OntologyDirective\x12\x0f\n\x07species\x18\x06 \x01(\t\x12I\n\x0cgenome_build\x18\x07 \x01(\x0b\x32\x33.nucleus.genomics.v1.GffHeader.GenomeBuildDirective\x1a \n\x11OntologyDirective\x12\x0b\n\x03uri\x18\x01 \x01(\t\x1a\x34\n\x14GenomeBuildDirective\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x12\n\x10GffReaderOptions\"\x12\n\x10GffWriterOptionsb\x06proto3')
  ,
  dependencies=[nucleus_dot_protos_dot_range__pb2.DESCRIPTOR,])



_GFFRECORD_STRAND = _descriptor.EnumDescriptor(
  name='Strand',
  full_name='nucleus.genomics.v1.GffRecord.Strand',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED_STRAND', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORWARD_STRAND', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REVERSE_STRAND', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=368,
  serialized_end=440,
)
_sym_db.RegisterEnumDescriptor(_GFFRECORD_STRAND)


_GFFRECORD_ATTRIBUTESENTRY = _descriptor.Descriptor(
  name='AttributesEntry',
  full_name='nucleus.genomics.v1.GffRecord.AttributesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.GffRecord.AttributesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.GffRecord.AttributesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=317,
  serialized_end=366,
)

_GFFRECORD = _descriptor.Descriptor(
  name='GffRecord',
  full_name='nucleus.genomics.v1.GffRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='range', full_name='nucleus.genomics.v1.GffRecord.range', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source', full_name='nucleus.genomics.v1.GffRecord.source', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='nucleus.genomics.v1.GffRecord.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='score', full_name='nucleus.genomics.v1.GffRecord.score', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='strand', full_name='nucleus.genomics.v1.GffRecord.strand', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='phase', full_name='nucleus.genomics.v1.GffRecord.phase', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attributes', full_name='nucleus.genomics.v1.GffRecord.attributes', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GFFRECORD_ATTRIBUTESENTRY, ],
  enum_types=[
    _GFFRECORD_STRAND,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=78,
  serialized_end=440,
)


_GFFHEADER_ONTOLOGYDIRECTIVE = _descriptor.Descriptor(
  name='OntologyDirective',
  full_name='nucleus.genomics.v1.GffHeader.OntologyDirective',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='nucleus.genomics.v1.GffHeader.OntologyDirective.uri', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=858,
  serialized_end=890,
)

_GFFHEADER_GENOMEBUILDDIRECTIVE = _descriptor.Descriptor(
  name='GenomeBuildDirective',
  full_name='nucleus.genomics.v1.GffHeader.GenomeBuildDirective',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='nucleus.genomics.v1.GffHeader.GenomeBuildDirective.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='nucleus.genomics.v1.GffHeader.GenomeBuildDirective.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=892,
  serialized_end=944,
)

_GFFHEADER = _descriptor.Descriptor(
  name='GffHeader',
  full_name='nucleus.genomics.v1.GffHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gff_version', full_name='nucleus.genomics.v1.GffHeader.gff_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequence_regions', full_name='nucleus.genomics.v1.GffHeader.sequence_regions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='feature_ontologies', full_name='nucleus.genomics.v1.GffHeader.feature_ontologies', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attribute_ontologies', full_name='nucleus.genomics.v1.GffHeader.attribute_ontologies', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source_ontologies', full_name='nucleus.genomics.v1.GffHeader.source_ontologies', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='species', full_name='nucleus.genomics.v1.GffHeader.species', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='genome_build', full_name='nucleus.genomics.v1.GffHeader.genome_build', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GFFHEADER_ONTOLOGYDIRECTIVE, _GFFHEADER_GENOMEBUILDDIRECTIVE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=443,
  serialized_end=944,
)


_GFFREADEROPTIONS = _descriptor.Descriptor(
  name='GffReaderOptions',
  full_name='nucleus.genomics.v1.GffReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=946,
  serialized_end=964,
)


_GFFWRITEROPTIONS = _descriptor.Descriptor(
  name='GffWriterOptions',
  full_name='nucleus.genomics.v1.GffWriterOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=966,
  serialized_end=984,
)

_GFFRECORD_ATTRIBUTESENTRY.containing_type = _GFFRECORD
_GFFRECORD.fields_by_name['range'].message_type = nucleus_dot_protos_dot_range__pb2._RANGE
_GFFRECORD.fields_by_name['strand'].enum_type = _GFFRECORD_STRAND
_GFFRECORD.fields_by_name['attributes'].message_type = _GFFRECORD_ATTRIBUTESENTRY
_GFFRECORD_STRAND.containing_type = _GFFRECORD
_GFFHEADER_ONTOLOGYDIRECTIVE.containing_type = _GFFHEADER
_GFFHEADER_GENOMEBUILDDIRECTIVE.containing_type = _GFFHEADER
_GFFHEADER.fields_by_name['sequence_regions'].message_type = nucleus_dot_protos_dot_range__pb2._RANGE
_GFFHEADER.fields_by_name['feature_ontologies'].message_type = _GFFHEADER_ONTOLOGYDIRECTIVE
_GFFHEADER.fields_by_name['attribute_ontologies'].message_type = _GFFHEADER_ONTOLOGYDIRECTIVE
_GFFHEADER.fields_by_name['source_ontologies'].message_type = _GFFHEADER_ONTOLOGYDIRECTIVE
_GFFHEADER.fields_by_name['genome_build'].message_type = _GFFHEADER_GENOMEBUILDDIRECTIVE
DESCRIPTOR.message_types_by_name['GffRecord'] = _GFFRECORD
DESCRIPTOR.message_types_by_name['GffHeader'] = _GFFHEADER
DESCRIPTOR.message_types_by_name['GffReaderOptions'] = _GFFREADEROPTIONS
DESCRIPTOR.message_types_by_name['GffWriterOptions'] = _GFFWRITEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GffRecord = _reflection.GeneratedProtocolMessageType('GffRecord', (_message.Message,), {

  'AttributesEntry' : _reflection.GeneratedProtocolMessageType('AttributesEntry', (_message.Message,), {
    'DESCRIPTOR' : _GFFRECORD_ATTRIBUTESENTRY,
    '__module__' : 'nucleus.protos.gff_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffRecord.AttributesEntry)
    })
  ,
  'DESCRIPTOR' : _GFFRECORD,
  '__module__' : 'nucleus.protos.gff_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffRecord)
  })
_sym_db.RegisterMessage(GffRecord)
_sym_db.RegisterMessage(GffRecord.AttributesEntry)

GffHeader = _reflection.GeneratedProtocolMessageType('GffHeader', (_message.Message,), {

  'OntologyDirective' : _reflection.GeneratedProtocolMessageType('OntologyDirective', (_message.Message,), {
    'DESCRIPTOR' : _GFFHEADER_ONTOLOGYDIRECTIVE,
    '__module__' : 'nucleus.protos.gff_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffHeader.OntologyDirective)
    })
  ,

  'GenomeBuildDirective' : _reflection.GeneratedProtocolMessageType('GenomeBuildDirective', (_message.Message,), {
    'DESCRIPTOR' : _GFFHEADER_GENOMEBUILDDIRECTIVE,
    '__module__' : 'nucleus.protos.gff_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffHeader.GenomeBuildDirective)
    })
  ,
  'DESCRIPTOR' : _GFFHEADER,
  '__module__' : 'nucleus.protos.gff_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffHeader)
  })
_sym_db.RegisterMessage(GffHeader)
_sym_db.RegisterMessage(GffHeader.OntologyDirective)
_sym_db.RegisterMessage(GffHeader.GenomeBuildDirective)

GffReaderOptions = _reflection.GeneratedProtocolMessageType('GffReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _GFFREADEROPTIONS,
  '__module__' : 'nucleus.protos.gff_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffReaderOptions)
  })
_sym_db.RegisterMessage(GffReaderOptions)

GffWriterOptions = _reflection.GeneratedProtocolMessageType('GffWriterOptions', (_message.Message,), {
  'DESCRIPTOR' : _GFFWRITEROPTIONS,
  '__module__' : 'nucleus.protos.gff_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.GffWriterOptions)
  })
_sym_db.RegisterMessage(GffWriterOptions)


_GFFRECORD_ATTRIBUTESENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading and writing VCF files.

The VCF format is described at
https://samtools.github.io/hts-specs/VCFv4.3.pdf

API for reading:

```python
from nucleus.io import vcf

with vcf.VcfReader(input_path) as reader:
  for variant in reader:
    print(variant)
```

API for writing:

```python
from nucleus.io import vcf

# variants is an iterable of nucleus.genomics.v1.Variant protocol buffers.
variants = ...

with vcf.VcfWriter(output_path, header=header) as writer:
  for variant in variants:
    writer.write(variant)
```

The class attempts to infer the file format (`TFRecord` vs VCF) from the file
path provided to the constructor.

1. For files that end with '.tfrecord' and '.tfrecord.gz' (a gzipped version),
  a `TFRecord` file is assumed and is attempted to be read or written.

2. For all other files, the VCF format will be used.

  VCF format used in writing is inferred from file paths:
    - ending in '.bcf.gz': BGZF compressed BCF format will be written;
    - ending in '.bcf': uncompressed BCF format will be written;
    - ending in '.gz' and not in '.bcf.gz': BGZP compressed VCF format will be
        written;
    - all other suffixes: uncompressed VCF format will be written.

  VCF format used in reading is inferred from the contents of the file.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import vcf_reader
from nucleus.io.python import vcf_writer
from nucleus.protos import variants_pb2
from nucleus.util import ranges
from nucleus.util import variant_utils
from nucleus.util import vcf_constants


def _create_get_fn_cache(fields):
  """Returns a dictionary from field to a callable that extracts its value."""
  return {
      field.id: vcf_constants.create_get_fn(field.type, field.number)
      for field in fields
  }


def _create_set_fn_cache(fields):
  """Returns a dictionary from field to a callable that sets its value."""
  return {field.id: vcf_constants.SET_FN_LOOKUP[field.type] for field in fields}


class VcfHeaderCache(object):
  """This class creates a cache of accessors to structured fields in Variants.

  The INFO and FORMAT fields within Variant protos are structured and typed,
  with types defined by the corresponding VCF header. This cache object provides
  provides {info,format}_field_{get,set}_fn functions that can be used to
  extract information from the structured Variant protos based on the types
  defined therein.

  NOTE: Users should not need to interact with this class at all. It is used
  by the variant_utils.{get,set}_info and variantcall_utils.{get,set}_format
  functions for interacting with the INFO and FORMAT fields in a Variant proto.
  """

  def __init__(self, header):
    """Initializer.

    Args:
      header: nucleus.genomics.v1.VcfHeader proto. Used to define the accessor
        functions needed.
    """
    if header is None:
      header = variants_pb2.VcfHeader()
    self._info_get_cache = _create_get_fn_cache(header.infos)
    self._info_set_cache = _create_set_fn_cache(header.infos)
    self._format_get_cache = _create_get_fn_cache(header.formats)
    self._format_set_cache = _create_set_fn_cache(header.formats)

  def info_field_get_fn(self, field_name):
    """Returns a callable that extracts the given INFO field based on its type.

    Args:
      field_name: str. The INFO field name of interest, e.g. 'AA', 'DB', 'AF'.

    Returns:
      A callable used to extract the given INFO field from a Variant proto.
    """
    return self._info_get_cache[field_name]

  def info_field_set_fn(self, field_name):
    """Returns a callable that sets the given INFO field based on its type."""
    return self._info_set_cache[field_name]

  def format_field_get_fn(self, field_name):
    """Returns a callable that gets the given FORMAT field based on its type."""
    return self._format_get_cache[field_name]

  def format_field_set_fn(self, field_name):
    """Returns a callable that sets the given FORMAT field based on its type."""
    return self._format_set_cache[field_name]


class NativeVcfReader(genomics_reader.GenomicsReader):
  """Class for reading from native VCF files.

  Most users will want to use VcfReader instead, because it dynamically
  dispatches between reading native VCF files and TFRecord files based
  on the filename's extensions.
  """

  def __init__(self,
               input_path,
               excluded_info_fields=None,
               excluded_format_fields=None,
               store_gl_and_pl_in_info_map=False,
               header=None):
    """Initializer for NativeVcfReader.

    Args:
      input_path: str. The path to the VCF file to read.
      excluded_info_fields: list(str). A list of INFO field IDs that should not
        be parsed into the Variants. If None, all INFO fields are included.
      excluded_format_fields: list(str). A list of FORMAT field IDs that should
        not be parsed into the Variants. If None, all FORMAT fields are
        included.
      store_gl_and_pl_in_info_map: bool. If True, the "GL" and "PL" FORMAT
        fields are stored in the VariantCall.info map rather than as top-level
        values in the VariantCall.genotype_likelihood field.
      header: If not None, specifies the variants_pb2.VcfHeader. The file at
        input_path must not contain any header information.
    """
    super(NativeVcfReader, self).__init__()

    options = variants_pb2.VcfReaderOptions(
        excluded_info_fields=excluded_info_fields,
        excluded_format_fields=excluded_format_fields,
        store_gl_and_pl_in_info_map=store_gl_and_pl_in_info_map)
    if header is not None:
      self._reader = vcf_reader.VcfReader.from_file_with_header(
          input_path.encode('utf8'), options, header)
    else:
      self._reader = vcf_reader.VcfReader.from_file(
          input_path.encode('utf8'), options)

    self.header = self._reader.header
    self.field_access_cache = VcfHeaderCache(self.header)

  def iterate(self):
    """Returns an iterable of Variant protos in the file."""
    return self._reader.iterate()

  def query(self, region):
    """Returns an iterator for going through variants in the region."""
    return self._reader.query(region)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)

  @property
  def c_reader(self):
    """Returns the underlying C++ reader."""
    return self._reader


class VcfReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading Variant protos from VCF or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeVcfReader(input_path, **kwargs)

  def _record_proto(self):
    return variants_pb2.Variant

  def _post_init_hook(self):
    # Initialize field_access_cache.  If we are dispatching to a
    # NativeVcfReader, we use its field_access_cache. Otherwise, we need to
    # create a new one.
    self.field_access_cache = getattr(
        self._reader, 'field_access_cache', VcfHeaderCache(self.header))

  @property
  def c_reader(self):
    """Returns the underlying C++ reader.

    Note that the C++ reader might be a VcfReader or it might be a
    TFRecordReader, depending on the input_path's extension.
    """
    return self._reader.c_reader


class NativeVcfWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native VCF files.

  Most users will want VcfWriter, which will write to either native VCF
  files or TFRecords files, based on the output filename's extensions.
  """

  def __init__(self,
               output_path,
               header=None,
               round_qualities=False,
               excluded_info_fields=None,
               excluded_format_fields=None,
               retrieve_gl_and_pl_from_info_map=False,
               exclude_header=False):
    """Initializer for NativeVcfWriter.

    Args:
      output_path: str. The path to which to write the VCF file.
      header: nucleus.genomics.v1.VcfHeader. The header that defines all
        information germane to the constituent variants. This includes contigs,
        FILTER fields, INFO fields, FORMAT fields, samples, and all other
        structured and unstructured header lines.
      round_qualities: bool. If True, the QUAL field is rounded to one point
        past the decimal.
      excluded_info_fields: list(str). A list of INFO field IDs that should not
        be written to the output. If None, all INFO fields are included.
      excluded_format_fields: list(str). A list of FORMAT field IDs that should
        not be written to the output. If None, all FORMAT fields are included.
      retrieve_gl_and_pl_from_info_map: bool. If True, the "GL" and "PL" FORMAT
        fields are retrieved from the VariantCall.info map rather than from the
        top-level value in the VariantCall.genotype_likelihood field.
      exclude_header: bool. If True, write a headerless VCF.
    """
    super(NativeVcfWriter, self).__init__()

    if header is None:
      header = variants_pb2.VcfHeader()
    writer_options = variants_pb2.VcfWriterOptions(
        round_qual_values=round_qualities,
        excluded_info_fields=excluded_info_fields,
        excluded_format_fields=excluded_format_fields,
        retrieve_gl_and_pl_from_info_map=retrieve_gl_and_pl_from_info_map,
        exclude_header=exclude_header,
    )
    self._writer = vcf_writer.VcfWriter.to_file(output_path, header,
                                                writer_options)
    self.field_access_cache = VcfHeaderCache(header)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class VcfWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing Variant protos to VCF or TFRecord files."""

  def _native_writer(self,
                     output_path,
                     header,
                     round_qualities=False,
                     excluded_info_fields=None,
                     excluded_format_fields=None,
                     retrieve_gl_and_pl_from_info_map=False,
                     exclude_header=False):
    return NativeVcfWriter(
        output_path,
        header=header,
        round_qualities=round_qualities,
        excluded_info_fields=excluded_info_fields,
        excluded_format_fields=excluded_format_fields,
        retrieve_gl_and_pl_from_info_map=retrieve_gl_and_pl_from_info_map,
        exclude_header=exclude_header)

  def _post_init_hook(self):
    # Initialize field_access_cache.  If we are dispatching to a
    # NativeVcfWriter, we use its field_access_cache.  Otherwise, we
    # need to create a new one.
    self.field_access_cache = getattr(
        self._writer, 'field_access_cache', VcfHeaderCache(self.header))


class InMemoryVcfReader(genomics_reader.GenomicsReader):
  """Class for "reading" Variant protos from an in-memory cache of variants.

  ```python
  from nucleus.io import vcf
  from nucleus.protos import variants_pb2

  variants = [... Variant protos ...]
  header = variants_pb2.VcfHeader()

  with vcf.InMemoryVcfReader(variants, header) as reader:
    for variant in reader:
      print(variant)
  ```

  This class accepts a collection of variants and optionally a header and
  provides all of the standard API functions of VcfReader but instead of
  fetching variants from a file the variants are queried from an in-memory cache
  of variant protos.

  Note that the input variants provided to this class aren't checked in any way,
  and their ordering determines the order of variants emitted by this class for
  the iterate() and query() operations. This is intentional, to make this class
  easy to use for testing where you often want to use less-than-perfectly formed
  inputs. In order to fully meet the contract of a standard VcfReader, variants
  should be sorted by their contig ordering and then by their start and finally
  by their ends.

  Implementation note:
    The current implementation will be very slow for query() if the provided
    cache of variants is large, as we do a O(n) search to collect all of the
    overlapping variants for each query. There are several straightforward
    optimizations to do if we need/want to scale this up. (a) sort the variants
    and use a binary search to find overlapping variants (b) partition the
    variants by contig, so we have dict[contig] => [variants on contig], which
    allows us to completely avoid considering any variants on any other contigs.
    Neither of these optimizations are worth it if len(variants) is small, but
    it may be worth considering if we want to use this functionality with a
    large number of variants.
  """

  def __init__(self, variants, header=None):
    """Creates a VCFReader backed by a collection of variants.

    Args:
      variants: list of nucleus.genomics.v1.Variant protos we will "read"
        from.
      header: a VCFHeader object to provide as a result to calls to self.header,
        or None, indicating that we don't have a header associated with this
        reader.
    """
    super(InMemoryVcfReader, self).__init__()
    self.variants = list(variants)
    self.header = header

  def iterate(self):
    return iter(self.variants)

  def query(self, region):
    return iter(
        variant for variant in self.variants
        if ranges.ranges_overlap(variant_utils.variant_range(variant), region)
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/fasta.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nucleus.protos import range_pb2 as nucleus_dot_protos_dot_range__pb2
from nucleus.protos import reference_pb2 as nucleus_dot_protos_dot_reference__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/fasta.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1anucleus/protos/fasta.proto\x12\x13nucleus.genomics.v1\x1a\x1anucleus/protos/range.proto\x1a\x1enucleus/protos/reference.proto\"\x8d\x01\n\x0b\x46\x61staRecord\x12\x0f\n\x07\x64\x65\x66line\x18\x01 \x01(\t\x12/\n\x06\x63ontig\x18\x02 \x01(\x0b\x32\x1f.nucleus.genomics.v1.ContigInfo\x12*\n\x06region\x18\x03 \x01(\x0b\x32\x1a.nucleus.genomics.v1.Range\x12\x10\n\x08sequence\x18\x04 \x01(\t\"\xde\x01\n\x12\x46\x61staReaderOptions\x12\x16\n\x0ekeep_true_case\x18\x01 \x01(\x08\x12\x10\n\x08\x61lphabet\x18\x02 \x01(\t\x12O\n\x0f\x64\x65\x66line_parsing\x18\x03 \x01(\x0e\x32\x36.nucleus.genomics.v1.FastaReaderOptions.DeflineParsing\x12 \n\x18include_range_in_records\x18\x04 \x01(\x08\"+\n\x0e\x44\x65\x66lineParsing\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0b\x43ONTIG_INFO\x10\x01\"\x14\n\x12\x46\x61staWriterOptionsb\x06proto3')
  ,
  dependencies=[nucleus_dot_protos_dot_range__pb2.DESCRIPTOR,nucleus_dot_protos_dot_reference__pb2.DESCRIPTOR,])



_FASTAREADEROPTIONS_DEFLINEPARSING = _descriptor.EnumDescriptor(
  name='DeflineParsing',
  full_name='nucleus.genomics.v1.FastaReaderOptions.DeflineParsing',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONTIG_INFO', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=435,
  serialized_end=478,
)
_sym_db.RegisterEnumDescriptor(_FASTAREADEROPTIONS_DEFLINEPARSING)


_FASTARECORD = _descriptor.Descriptor(
  name='FastaRecord',
  full_name='nucleus.genomics.v1.FastaRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defline', full_name='nucleus.genomics.v1.FastaRecord.defline', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contig', full_name='nucleus.genomics.v1.FastaRecord.contig', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='region', full_name='nucleus.genomics.v1.FastaRecord.region', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequence', full_name='nucleus.genomics.v1.FastaRecord.sequence', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=112,
  serialized_end=253,
)


_FASTAREADEROPTIONS = _descriptor.Descriptor(
  name='FastaReaderOptions',
  full_name='nucleus.genomics.v1.FastaReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='keep_true_case', full_name='nucleus.genomics.v1.FastaReaderOptions.keep_true_case', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alphabet', full_name='nucleus.genomics.v1.FastaReaderOptions.alphabet', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='defline_parsing', full_name='nucleus.genomics.v1.FastaReaderOptions.defline_parsing', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='include_range_in_records', full_name='nucleus.genomics.v1.FastaReaderOptions.include_range_in_records', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _FASTAREADEROPTIONS_DEFLINEPARSING,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=256,
  serialized_end=478,
)


_FASTAWRITEROPTIONS = _descriptor.Descriptor(
  name='FastaWriterOptions',
  full_name='nucleus.genomics.v1.FastaWriterOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=480,
  serialized_end=500,
)

_FASTARECORD.fields_by_name['contig'].message_type = nucleus_dot_protos_dot_reference__pb2._CONTIGINFO
_FASTARECORD.fields_by_name['region'].message_type = nucleus_dot_protos_dot_range__pb2._RANGE
_FASTAREADEROPTIONS.fields_by_name['defline_parsing'].enum_type = _FASTAREADEROPTIONS_DEFLINEPARSING
_FASTAREADEROPTIONS_DEFLINEPARSING.containing_type = _FASTAREADEROPTIONS
DESCRIPTOR.message_types_by_name['FastaRecord'] = _FASTARECORD
DESCRIPTOR.message_types_by_name['FastaReaderOptions'] = _FASTAREADEROPTIONS
DESCRIPTOR.message_types_by_name['FastaWriterOptions'] = _FASTAWRITEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FastaRecord = _reflection.GeneratedProtocolMessageType('FastaRecord', (_message.Message,), {
  'DESCRIPTOR' : _FASTARECORD,
  '__module__' : 'nucleus.protos.fasta_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastaRecord)
  })
_sym_db.RegisterMessage(FastaRecord)

FastaReaderOptions = _reflection.GeneratedProtocolMessageType('FastaReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _FASTAREADEROPTIONS,
  '__module__' : 'nucleus.protos.fasta_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastaReaderOptions)
  })
_sym_db.RegisterMessage(FastaReaderOptions)

FastaWriterOptions = _reflection.GeneratedProtocolMessageType('FastaWriterOptions', (_message.Message,), {
  'DESCRIPTOR' : _FASTAWRITEROPTIONS,
  '__module__' : 'nucleus.protos.fasta_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastaWriterOptions)
  })
_sym_db.RegisterMessage(FastaWriterOptions)


# @@protoc_insertion_point(module_scope)
